<?php

namespace Drupal\klaro\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\klaro\Utility\KlaroHelper;
use InvalidArgumentException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Routing\Exception\RouteNotFoundException;

/**
 * Provides an add form for a Klaro! app.
 *
 * @internal
 */
class KlaroAppForm extends EntityForm {

  /**
   * The Klaro! helper service.
   *
   * @var \Drupal\klaro\Utility\KlaroHelper
   */
  protected $klaroHelper;

  /**
   * Constructs an ExampleForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entityTypeManager.
   * @param \Drupal\klaro\Utility\KlaroHelper $klaro_helper
   *   The Klaro! helper service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, KlaroHelper $klaro_helper) {
    $this->entityTypeManager = $entityTypeManager;
    $this->klaroHelper = $klaro_helper;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('klaro.helper')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\klaro\KlaroAppInterface $app */
    $app = $this->entity;
    $form = parent::form($form, $form_state);
    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $app->label(),
      '#description' => $this->t("The label for the App. The label will appear on the <em>Klaro! consent manager</em> modal."),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $app->id(),
      '#description' => $this->t('A unique machine-readable name for this Klaro! app.'),
      '#maxlength' => 32,
      '#machine_name' => [
        'exists' => [$this, 'exist'],
        'source' => ['label'],
      ],
      '#disabled' => !$app->isNew(),
      '#required' => TRUE,
    ];

    $form['description'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Description'),
      '#maxlength' => 255,
      '#default_value' => $app->description(),
      '#description' => $this->t('Describe this app. The text will appear on the <em>Klaro! consent manager</em> modal.'),
      '#required' => TRUE,
    ];

    $form['status'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => $app->isNew() ? FALSE : $app->status(),
    ];

    $form['klaro'] = [
      '#type' => 'vertical_tabs',
      '#title' => $this->t('Settings'),
    ];

    $form['info'] = [
      '#type' => 'details',
      '#title' => $this->t('Information'),
      '#group' => 'klaro',
      '#summary_attributes' => ['id' => ['checked']],
      '#weight' => 2,
    ];

    $form['info']['info_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Info url'),
      '#description' => [
        '#theme' => 'item_list',
        '#items' => [
          $this->t('The url where people can get general information about this app.'),
          $this->t('You can enter an internal path such as %path or an external URL such as %url', [
            '%path' => 'internal:/node/add',
            '%url' => 'http://example.com',
          ]),
        ],
      ],
      '#default_value' => $app->infoUrl(),
      '#placeholder' => 'https://',
      '#size' => 255,
      '#element_validate' => [
        [get_class($this), 'validateUrl'],
      ],
    ];

    $form['info']['privacy_policy_url'] = [
      '#type' => 'url',
      '#title' => $this->t('Privacy policy url'),
      '#description' => [
        '#theme' => 'item_list',
        '#items' => [
          $this->t('The url where people can read the privacy policies of using this app.'),
          $this->t('You can enter an internal path such as %path or an external URL such as %url', [
            '%path' => 'internal:/node/add',
            '%url' => 'http://example.com',
          ]),
        ],
      ],
      '#default_value' => $app->privacyPolicyUrl(),
      '#placeholder' => 'internal:/<front>',
      '#size' => 255,
      '#element_validate' => [
        [get_class($this), 'validateUrl'],
      ],
    ];

    $form['options'] = [
      '#type' => 'details',
      '#title' => $this->t('Options'),
      '#group' => 'klaro',
      '#required' => TRUE,
      '#weight' => 3,
    ];

    $form['options']['purposes'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Purposes'),
      '#description' => $this->t('Which purposes does this app match best? This information will appear on the Klaro! consent manager modal.'),
      '#required' => TRUE,
      '#options' => $this->klaroHelper->optionPurposes(),
      '#default_value' => $app->purposes(),
    ];

    $form['options']['default'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Default'),
      '#description' => $this->t('The default state of the this Klaro! app. Will appear already enabled on the <em>Klaro! consent manager</em> modal.'),
      '#default_value' => $app->isDefault(),
    ];

    $form['options']['required'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Always required'),
      '#description' => $this->t('The use will not be able to disable it on <em>Klaro! consent manager</em> modal.'),
      '#default_value' => $app->isRequired(),
    ];

    $form['options']['opt_out'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Opt out'),
      '#description' => $this->t('This Klaro! app will be loaded even before the user gave explicit consent.'),
      '#default_value' => $app->isOptOut(),
    ];

    $form['options']['only_once'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Only once'),
      '#description' => $this->t('The Klaro! app will only be executed once. Regardless how often the user toggles it on and off.'),
      '#default_value' => $app->isOnlyOnce(),
    ];

    $form['advanced'] = [
      '#type' => 'details',
      '#title' => $this->t('Advanced'),
      '#group' => 'klaro',
      '#weight' => 4,
      '#tree' => FALSE,
    ];

    $form['advanced']['cookies'] = [
      '#type' => 'details',
      '#title' => $this->t('Cookies'),
      '#open' => TRUE,
      '#tree' => TRUE,
      '#prefix' => '<div id="cookies-wrapper">',
      '#suffix' => '</div>',
      '#description' => [
        '#markup' => $this->t('Klaro! only deletes cookies of this app if matching the cookie name by regex AND exact matching the cookie path AND exact matching the cookie domain.'),
        'options' => [
          '#theme' => 'item_list',
          '#items' => [
            $this->t('Use any regular expression to identifiy the cookie value.'),
            $this->t('Leave the path empty to use "/".'),
            $this->t('Leave domain empty to use the browsers location.host.'),
          ],
        ],
      ],
      '#description_display' => 'before',
      '#weight' => 0,
    ];

    $cookies = $form_state->get('cookies') ?: ($form_state->getTriggeringElement() ? [] : $app->cookies());
    $form_state->set('cookies', $cookies);

    foreach ($cookies as $delta => $cookie) {
      $form['advanced']['cookies'][$delta] = [
        '#type' => 'container',
        '#open' => TRUE,
        '#attributes' => [
          'class' => ['container-inline'],
          'style' => 'border: 1px dotted;padding: 0px 10px',
        ],
      ];

      $form['advanced']['cookies'][$delta]['regex'] = [
        '#type' => 'textfield',
        '#field_prefix' => '/',
        '#field_suffix' => '/i',
        '#maxlength' => 50,
        '#size' => 20,
        '#title' => $this->t('Regex'),
        '#default_value' => isset($cookie['regex']) ? $cookie['regex'] : '',
      ];

      $form['advanced']['cookies'][$delta]['path'] = [
        '#type' => 'textfield',
        '#maxlength' => 50,
        '#size' => 20,
        '#title' => $this->t('Path'),
        '#default_value' => isset($cookie['path']) ? $cookie['path'] : '',
        '#description_display' => 'before',
      ];

      $form['advanced']['cookies'][$delta]['domain'] = [
        '#type' => 'textfield',
        '#maxlength' => 50,
        '#size' => 20,
        '#title' => $this->t('Domain'),
        '#default_value' => isset($cookie['domain']) ? $cookie['domain'] : '',
        '#description_display' => 'before',
      ];

      $form['advanced']['cookies'][$delta]['delete'] = [
        '#type' => 'submit',
        '#name' => "delete[$delta]",
        '#value' => 'X',
        '#submit' => ['::ajaxCookieSubmit'],
        '#ajax' => [
          'callback' => '::returnCookies',
          'wrapper' => 'cookies-wrapper',
        ],
        '#attributes' => [
          'title' => $this->t('Delete'),
          'class' => [
            'button-action',
            'button--small',
          ],
        ],
        '#button_type' => 'primary',
        '#limit_validation_errors' => [],
      ];
    }

    $form['advanced']['cookies']['actions'] = [
      '#type' => 'actions',
    ];

    $form['advanced']['cookies']['actions']['add'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add'),
      '#submit' => ['::ajaxCookieSubmit'],
      '#ajax' => [
        'callback' => '::returnCookies',
        'wrapper' => 'cookies-wrapper',
      ],
      '#attributes' => [
        'title' => $this->t('Add another cookie information'),
        'class' => [
          'button-action',
          'button--small',
        ],
      ],
      '#button_type' => 'primary',
      '#limit_validation_errors' => [],
    ];

    $form['advanced']['js'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Javascripts'),
      '#description' => [
        '#markup' => $this->t('Klaro! can only take control over javascripts if they are first injected into the DOM as text/plain and containing several data-attributes. To get those javascripts manipulated in that way, Klaro! needs a filepath or filename to identify them.'),
        'options' => [
          '#theme' => 'item_list',
          '#items' => [
            $this->t('Enter one javascript per line'),
          ],
        ],
      ],
      '#default_value' => implode("\n", $app->javascripts()),
    ];

    $form['advanced']['att'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Attachments'),
      '#description' => [
        '#markup' => $this->t('Some javascripts are added as script tag via drupal "attach" including a unique identifier. If Klaro! should take control over these scripts, enter the IDs here.'),
        'options' => [
          '#theme' => 'item_list',
          '#items' => [
            $this->t('Enter one ID per line'),
          ],
        ],
      ],
      '#default_value' => implode("\n", $app->attachments()),
    ];

    $form['weight'] = [
      '#type' => 'number',
      '#min' => -99,
      '#max' => 99,
      '#default_value' => $app->weight(),
      '#title' => $this->t('Weight'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\klaro\KlaroAppInterface $app */
    $app = $this->entity;
    $cookies = $form_state->getValue('cookies');
    // Lets filter out action and empty cookie infos.
    $filtered_cookies = [];
    foreach ($cookies as $i => $info) {
      if (
        $i === 'action' ||
        (empty($info['regex']) && empty($info['path']) && empty($info['domain']))
      ) {
        continue;
      }
      $filtered_cookies[] = $info;
    }
    $app->setCookies($filtered_cookies);
    $app->setPurposes(array_keys(array_filter($form_state->getValue('purposes'))));
    $app->setJavaScripts(array_filter(array_map('trim', explode("\n", $form_state->getValue('js')))));
    $app->setAttachments(array_filter(array_map('trim', explode("\n", $form_state->getValue('att')))));

    $status = $app->save();

    if ($status === SAVED_NEW) {
      $this->messenger()->addMessage($this->t('The Klaro! app %label has been created.', [
        '%label' => $app->label(),
      ]));
    }
    else {
      $this->messenger()->addMessage($this->t('The Klaro! app %label has been updated.', [
        '%label' => $app->label(),
      ]));
    }

    $form_state->setRedirect('entity.klaro_app.collection');
  }

  /**
   * Helper function to check whether an Klaro! app configuration entity exists.
   */
  public function exist($id) {
    $entity = $this->entityTypeManager->getStorage('klaro_app')->getQuery()
      ->condition('id', $id)
      ->execute();
    return (bool) $entity;
  }

  /**
   * Returns the cookies form elements.
   */
  public function returnCookies(array &$form, FormStateInterface $form_state) {
    return $form['advanced']['cookies'];
  }

  /**
   * Adds / removes cookies to / from the form_state.
   */
  public function ajaxCookieSubmit(array &$form, FormStateInterface $form_state) {
    $cookies = $form_state->get('cookies');
    $parents = $form_state->getTriggeringElement()['#parents'];

    switch ($parents[2]) {
      case 'add':
        $cookies[] = [
          'regex' => '',
          'path' => '',
          'domain' => '',
        ];
        break;

      case 'delete':
        unset($cookies[$parents[1]]);
        break;
    }

    $form_state->set('cookies', $cookies);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Form element validation handler replacement for #type 'url'.
   */
  public static function validateUrl(&$element, FormStateInterface $form_state) {
    $value = trim($element['#value']);
    $form_state->setValueForElement($element, $value);

    if ($value !== '' && !UrlHelper::isValid($value, TRUE)) {
      // Value is not a valid absolute url.
      try {
        // Make sure the uri looks sensible.
        $url = Url::fromUri($value);
        // Make sure we can actually build a link with this uri.
        $url_string = $url->toString();
      }
      // @todo Provide translated messages for the various error states.
      catch (InvalidArgumentException $e) {
        $form_state->setError($element, t('The URL %url is not valid: @error', [
          '%url' => $value,
          '@error' => $e->getMessage(),
        ]));
      }
      catch (RouteNotFoundException $e) {
        $form_state->setError($element, t('The specified route does not exist.'));
      }
    }
  }

}
