<?php

namespace Drupal\klaro\Form;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use InvalidArgumentException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Routing\Exception\RouteNotFoundException;

/**
 * General settings form for the Klaro! consent manager.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * Klaro! settings for the drupal integration.
   */
  const SETTINGS = 'klaro.settings';

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'klaro_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      static::SETTINGS,
    ];
  }

  /**
   * The cache data service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheData;

  /**
   * Constructs a \Drupal\system\ConfigFormBase object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_data_service
   *   The cache data service.
   */
  public function __construct(ConfigFactoryInterface $config_factory, CacheBackendInterface $cache_data_service) {
    parent::__construct($config_factory);
    $this->cacheData = $cache_data_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('cache.data')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config(static::SETTINGS);

    $form['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Klaro! enabled'),
      '#description' => $this->t('The global state of the Klaro! implementation to the front end.'),
      '#default_value' => $config->get('enabled'),
    ];

    $form['klaro'] = [
      '#type' => 'vertical_tabs',
      '#title' => $this->t('Library settings'),
    ];

    // General settings.
    $form['general_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('General'),
      '#group' => 'klaro',
      '#weight' => 1,
    ];

    $form['general_settings']['element_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Element ID'),
      '#description' => $this->t('You can customize the ID of the DIV element that Klaro! will create when starting up.'),
      '#required' => TRUE,
      '#default_value' => $config->get('library.element_id'),
      '#element_validate' => [
        [get_class($this), 'validateElementId'],
      ],
    ];

    $form['general_settings']['privacy_policy'] = [
      '#type' => 'url',
      '#title' => $this->t('Privacy policy Url'),
      '#description' => $this->t('You can enter an internal path such as %path or an external URL such as %url', [
        '%path' => 'internal:/node/add',
        '%url' => 'http://example.com',
      ]),
      '#required' => TRUE,
      '#default_value' => $config->get('library.privacy_policy'),
      '#size' => 255,
      '#element_validate' => [
        [get_class($this), 'validateUrl'],
      ],
    ];

    // Storage settings.
    $form['storage_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Storage settings'),
      '#group' => 'klaro',
      '#weight' => 2,
    ];

    $form['storage_settings']['storage_method'] = [
      '#type' => 'select',
      '#title' => $this->t('Storage method'),
      '#description' => $this->t('How Klaro! should store the preferences of the user.'),
      '#required' => TRUE,
      '#options' => [
        'cookie' => $this->t('Cookie'),
        'localStorage' => $this->t('Local storage'),
      ],
      '#default_value' => $config->get('library.storage_method'),
    ];

    $form['storage_settings']['cookie_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Cookie name'),
      '#description' => $this->t('You can customize the name of the cookie that Klaro! uses for storing user consent decisions.'),
      '#default_value' => $config->get('library.cookie_name'),
    ];

    $form['storage_settings']['cookie_expires_after_days'] = [
      '#type' => 'number',
      '#title' => $this->t('Cookie expires after'),
      '#description' => $this->t('You can also set a custom expiration time for the Klaro! cookie.'),
      '#min' => 0,
      '#max' => 365,
      '#step' => 1,
      '#field_suffix' => $this->t('days'),
      '#default_value' => $config->get('library.cookie_expires_after_days'),
    ];

    $form['storage_settings']['cookie_domain'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Cookie domain'),
      '#description' => $this->t('You can change to cookie domain for the consent manager itself. Use this if you want to get consent once for multiple matching domains. If undefined, Klaro! will use the current domain.'),
      '#default_value' => $config->get('library.cookie_domain'),
    ];

    // Consent settings.
    $form['options'] = [
      '#type' => 'details',
      '#title' => $this->t('Options'),
      '#description' => $this->t('Configure options for the Klaro! consent manager modal.'),
      '#group' => 'klaro',
      '#weight' => 3,
    ];

    $form['options']['default'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Default state'),
      '#description' => $this->t('Defines the default state for applications.'),
      '#default_value' => $config->get('library.default'),
    ];

    $form['options']['must_consent'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Must consent'),
      '#description' => $this->t('If set, Klaro! will directly display the consent manager modal and not allow the user to close it before having actively consented or declines the use of third-party apps.'),
      '#default_value' => $config->get('library.must_consent'),
    ];

    $form['options']['accept_all'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Accept all'),
      '#description' => $this->t('Show "accept all" to accept all apps instead of "ok" that only accepts required and "default: true" apps'),
      '#default_value' => $config->get('library.accept_all'),
    ];

    $form['options']['hide_decline_all'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show decline button'),
      '#description' => $this->t('Shows the "decline" button on the Klaro! consent manager modal.'),
      '#default_value' => !$config->get('library.hide_decline_all'),
    ];

    // Consent settings.
    $form['advanced'] = [
      '#type' => 'details',
      '#title' => $this->t('Advanced'),
      '#description' => $this->t('Configure advanced settings.'),
      '#group' => 'klaro',
      '#weight' => 4,
    ];

    $form['advanced']['deletable_cookie_domains'] = [
      '#type' => 'textarea',
      '#rows' => 5,
      '#title' => $this->t('Matching cookie domains'),
      '#description' => $this->t('Enter one domain per line for cookie deletion. Leave empty to delete the current domain of the visitor only.'),
      '#default_value' => implode("\n", $config->get('deletable_cookie_domains')),
    ];

    $form['advanced']['klaro_library'] = [
      '#type' => 'select',
      '#title' => $this->t('Klaro! library'),
      '#description' => [
        '#markup' => $this->t('Which Klaro! library should be used?'),
        'options' => [
          '#theme' => 'item_list',
          '#items' => [
            $this->t('<em>Klaro! (default)</em>: Klaro! will load the stylesheets by javascript. The stylesheets will <b>not</b> be preprocessable by drupal.'),
            $this->t('<em>Klaro! (no css)</em>: Klaro! will not load the stylesheets. The stylesheets will be loaded by drupal and will be preprocessable.'),
          ],
        ],
      ],
      '#options' => [
        'klaro' => $this->t('Klaro! (default)'),
        'klaro_no_css' => $this->t('Klaro! (no css)'),
      ],
      '#required' => TRUE,
      '#default_value' => $config->get('klaro_library'),
    ];

    $form['advanced']['process_descriptions'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Process descriptions'),
      '#description' => $this->t('If enabled, all Klaro! app descriptions will be processed. As for now they will get extended by the privacy policy url and the info url.'),
      '#default_value' => $config->get('process_descriptions'),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $deletabel_cookie_domains = array_filter(array_map('trim', explode("\n", $form_state->getValue('deletable_cookie_domains'))));
    $this->configFactory->getEditable(static::SETTINGS)
      ->set('library.element_id', $form_state->getValue('element_id'))
      ->set('library.storage_method', $form_state->getValue('storage_method'))
      ->set('library.cookie_name', $form_state->getValue('cookie_name'))
      ->set('library.cookie_expires_after_days', $form_state->getValue('cookie_expires_after_days'))
      ->set('library.cookie_domain', $form_state->getValue('cookie_domain'))
      ->set('library.privacy_policy', $form_state->getValue('privacy_policy'))
      ->set('library.default', $form_state->getValue('default'))
      ->set('library.must_consent', $form_state->getValue('must_consent'))
      ->set('library.accept_all', $form_state->getValue('accept_all'))
      ->set('library.hide_decline_all', !$form_state->getValue('hide_decline_all'))
      ->set('deletable_cookie_domains', $deletabel_cookie_domains)
      ->set('enabled', $form_state->getValue('enabled'))
      ->set('klaro_library', $form_state->getValue('klaro_library'))
      ->set('process_descriptions', $form_state->getValue('process_descriptions'))
      ->save();

    parent::submitForm($form, $form_state);
  }

  /**
   * Validates the string an reformats it to fit an html id.
   */
  public static function validateElementId(&$element, FormStateInterface $form_state) {
    $original = $element['#value'];
    $new = Html::cleanCssIdentifier($original);
    $form_state->setValueForElement($element, $new);

    if ($original !== $new) {
      \Drupal::messenger()->addWarning(t('Your original value %original has been replaced by a valid element ID: %new. Please check and save again.', [
        '%original' => $original,
        '%new' => $new,
      ]));
    }
  }

  /**
   * Form element validation handler replacement for #type 'url'.
   */
  public static function validateUrl(&$element, FormStateInterface $form_state) {
    $value = trim($element['#value']);
    $form_state->setValueForElement($element, $value);

    if ($value !== '' && !UrlHelper::isValid($value, TRUE)) {
      // Value is not a valid absolute url.
      try {
        // Make sure the uri looks sensible.
        $url = Url::fromUri($value);
        // Make sure we can actually build a link with this uri.
        $url_string = $url->toString();
      }
      // @todo Provide translated messages for the various error states.
      catch (InvalidArgumentException $e) {
        $form_state->setError($element, t('The URL %url is not valid: @error', [
          '%url' => $value,
          '@error' => $e->getMessage(),
        ]));
      }
      catch (RouteNotFoundException $e) {
        $form_state->setError($element, t('The specified route does not exist.'));
      }
    }
  }

}
