<?php

namespace Drupal\klaro\Utility;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\klaro\KlaroAppInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides helper methods for klaro.
 */
class KlaroHelper implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * The context key for Klaro! specific translatable strings.
   *
   * @var string
   */
  const TRANSLATION_CONTEXT = 'klaro';

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The current user service.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a EmailAction object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languages
   *   The language manager.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    LanguageManagerInterface $languages,
    AccountProxyInterface $current_user,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->configFactory = $config_factory;
    $this->languageManager = $languages;
    $this->currentUser = $current_user;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('language_manager'),
      $container->get('current_user'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Gets all available purposes.
   *
   * @return array
   *   The purposes.
   */
  public function getAvailablePurposes(): array {
    $purpose_storage = $this->entityTypeManager->getStorage('klaro_purpose');

    $query = $purpose_storage->getQuery();
    $query->sort('weight');

    $result = $query->execute();

    return $purpose_storage->loadMultiple($result);
  }

  /**
   * Returns if all necessary dependencies are satisfied to use klaro.
   *
   * If plugin_id is given an additionally check is performed if specified apps
   * are enabled. Will return FALSE if one of the apps are not enabled.
   *
   * @param array $app_ids
   *   The app ids to check.
   *
   * @return bool
   *   If has access in general or to a specific app.
   */
  public function hasAccess(array $app_ids = []): bool {
    if ($this->currentUser->hasPermission('bypass klaro')) {
      return FALSE;
    }

    $config = $this->configFactory->get('klaro.settings');

    if (!$config->get('enabled')) {
      return FALSE;
    }

    $enabled_apps = $this->getApps();

    if (empty($enabled_apps)) {
      return FALSE;
    }

    if (!empty($app_ids)) {
      $enabled_app_ids = array_map(function (KlaroAppInterface $a) {
        return $a->id();
      }, $enabled_apps);

      foreach ($app_ids as $app_id) {
        if (!in_array($app_id, $enabled_app_ids)) {
          return FALSE;
        }
      }
    }

    return TRUE;
  }

  /**
   * Get an array that should be used for drupalSettings.
   *
   * @return array
   *   The drupalSettings array.
   */
  public function processDrupalSettings(): array {
    $settings = [];
    $current_lang = $this->languageManager->getCurrentLanguage()->getId();
    $config = $this->configFactory->get('klaro.settings');
    $config_texts = $this->configFactory->get('klaro.texts');

    // Add library settings.
    $library_settings = $config->get('library');
    $library_settings['privacy_policy'] = Url::fromUri($config->get('library.privacy_policy'))->toString();
    $settings['config'] = static::snakeToCamel($library_settings);
    $deletabel_cookie_domains = $config->get('deletable_cookie_domains');
    // Add all apps to the config.
    foreach ($this->getApps() as $app) {
      $cookies = array_map('array_values', $app->cookies());
      foreach ($app->cookies() as $info) {
        foreach ($deletabel_cookie_domains as $del_domain) {
          $cookies[] = [$info['regex'], $info['path'], $del_domain];
        }
      }
      $settings['config']['apps'][] = [
        'name' => $app->id(),
        'default' => $app->isDefault(),
        'title' => $app->label(),
        'description' => $config->get('process_descriptions') ? $this->processDescription($app) : $app->description(),
        'purposes' => $app->purposes(),
        'cookies' => $cookies,
        'required' => $app->isRequired(),
        'optOut' => $app->isOptOut(),
        'onlyOnce' => $app->isOnlyOnce(),
      ];
    }

    $translations = $config_texts->get();
    if (!isset($translations['poweredBy']) || empty($translations['poweredBy'])) {
      $translations['poweredBy'] = '';
    }
    unset($translations['_core'], $translations['langcode']);
    $settings['config']['translations'][$current_lang] = $translations;
    $settings['config']['translations'][$current_lang]['purposes'] = $this->optionPurposes();

    return $settings;
  }

  /**
   * Convert under_score type array keys to camelCase type array keys.
   *
   * @param array $array
   *   The array to convert.
   *
   * @return array
   *   The array with converted keys.
   */
  private static function snakeToCamel(array $array): array {
    $finalArray = [];

    foreach ($array as $key => $value) {
      if (strpos($key, "_")) {
        $key = lcfirst(str_replace("_", "", ucwords($key, "_")));
      }
      if (!is_array($value)) {
        $finalArray[$key] = $value;
      }
      else {
        $finalArray[$key] = static::snakeToCamel($value);
      }
    }

    return $finalArray;
  }

  /**
   * Get all apps.
   *
   * @param bool $only_enabled
   *   If only enabled apps should be fetched.
   *
   * @return \Drupal\klaro\KlaroAppInterface[]
   *   The enabled apps.
   */
  public function getApps(bool $only_enabled = TRUE): array {
    $app_storage = $this->entityTypeManager->getStorage('klaro_app');

    $query = $app_storage->getQuery();
    $query->sort('weight');
    if ($only_enabled) {
      $query->condition('status', TRUE);
    }

    $result = $query->execute();

    return $app_storage->loadMultiple($result);
  }

  /**
   * Retrieves available purposes as an options array.
   *
   * @return array
   *   The purposes options.
   */
  public function optionPurposes(): array {
    $options = [];

    foreach ($this->getAvailablePurposes() as $purpose) {
      $options[$purpose->id()] = $purpose->label();
    }

    return $options;
  }

  /**
   * Creates the description of the app by combining several information.
   *
   * @param \Drupal\klaro\KlaroAppInterface $app
   *   The Klaro! app.
   *
   * @return string
   *   The processed description.
   */
  public function processDescription(KlaroAppInterface $app): string {
    $description = $app->description();

    $urls = [
      'privacy_policy' => $app->privacyPolicyUrl(),
      'info' => $app->infoUrl(),
    ];

    foreach (array_filter($urls) as $key => $url) {
      $description .= " - ";
      switch ($key) {
        case 'privacy_policy':
          $description .= $this->t('Privacy policy');
          break;

        case 'info':
          $description .= $this->t('Info');
          break;
      }
      $description .= ": {$url}";
    }

    return $description;
  }

}
